/*
 *  renderer - A simple implementation of polygon-based 3D algorithms.
 *  Copyright (C) 2004  Thanassis Tsiodras (ttsiodras@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __types_h__
#define __types_h__

#include <config.h>

#include <math.h>

// The main floating point type, used everywhere
// If you need more accuracy for some reason, use "double" (30% speed hit)
typedef float coord;

struct Vector3
{
    coord _x, _y, _z;

    Vector3(coord x=0, coord y=0, coord z=0)
	:
	_x(x), _y(y), _z(z) {}

    Vector3(const Vector3& rhs)
	:
	_x(rhs._x),
	_y(rhs._y),
	_z(rhs._z) {}

/* An abandoned experiment with intrinsics
 * (GCC proved to do just as good a job in optimizing with SSE)

        __m128 tmp = _mm_loadu_ps(&_x);
	__m128 l = _mm_mul_ps(tmp, tmp);
	l = _mm_add_ps(l, _mm_shuffle_ps(l, l, 0x4E));
	_mm_storeu_ps(&_x, _mm_mul_ps(tmp, _mm_rsqrt_ps(_mm_add_ps(l, _mm_shuffle_ps(l, l, 0x11)))));

 */
    void normalize()
    {
	coord norm = sqrt(_x*_x +_y*_y + _z*_z);
	_x /= norm;
	_y /= norm;
	_z /= norm;
    }
#ifdef CUDA_VERSION
    __device__ void normalizeCUDA()
    {
	coord norm = sqrt(_x*_x +_y*_y + _z*_z);
	_x /= norm;
	_y /= norm;
	_z /= norm;
    }

    __device__ coord length()
    {
	return sqrt(_x*_x +_y*_y + _z*_z);
    }

#endif
    Vector3& operator+=(const Vector3& rhs)
    {
	_x += rhs._x;
	_y += rhs._y;
	_z += rhs._z;
	return *this;
    }

    Vector3& operator-=(const Vector3& rhs)
    {
	_x -= rhs._x;
	_y -= rhs._y;
	_z -= rhs._z;
	return *this;
    }

    Vector3& operator*=(const coord& rhs)
    {
	_x *= rhs;
	_y *= rhs;
	_z *= rhs;
	return *this;
    }

    Vector3& operator/=(const coord& rhs)
    {
	_x /= rhs;
	_y /= rhs;
	_z /= rhs;
	return *this;
    }
};

// In the Ambient and Gouraud modes, we interpolate color per-pixel...
// We therefore need floating point components for the colors
struct Pixel {
    float _b;
    float _g;
    float _r;
    Pixel(unsigned char r=0.f, unsigned char g=0.f, unsigned char b=0.f)
	:
	_b(b), _g(g), _r(r) {}

    Pixel& operator+=(const Pixel& rhs) { _b += rhs._b; _g += rhs._g; _r += rhs._r; return *this; }
    Pixel& operator-=(const Pixel& rhs) { _b -= rhs._b; _g -= rhs._g; _r -= rhs._r; return *this; }
    Pixel& operator*=(const coord& rhs) {
	_b = rhs*_b;
	_g = rhs*_g;
	_r = rhs*_r;
	return *this;
    }
    Pixel& operator/=(const coord& rhs) {
	_b = _b/rhs;
	_g = _g/rhs;
	_r = _r/rhs;
	return *this;
    }
};

#endif
